#ifndef LIBHARKIO_CHUNK_H
#define LIBHARKIO_CHUNK_H

#include <cstring>
#include <iostream>
#include <fstream>
#include "common.h"

namespace harkio {

/**********************************************************************/
//! Chunk抽象基底クラス
/*!
 * 	Chunk階層に属する実際のデータは、このクラスの派生として実現する
 */
/**********************************************************************/
class Chunk{
public:
	//! checkVersion/checkVersionCommon の戻り値
	typedef enum {MISMATCH, SAFE_MATCH, PERFECT_MATCH} VERSION_CHECK;
protected:
	//! \name フィールドデータ
	//\{
	int size; 	//!< サイズ
	std::pair<short int, short int> version; //!< メジャーバージョン,マイナーバージョン
	//\}

	//! バージョンチェック用共通関数
	/*! バージョンチェックは必ず派生クラスのcheckVersion()関数から呼ばれるのでテンプレート化してある
	 *  直接この関数を呼ぶ事はない
	 */
	template<typename T> VERSION_CHECK checkVersionCommon(T* derive, typename T::VNAME required_vname) {
		std::vector<std::pair<short, short> >* vlist = derive->getVersions();
		std::pair<short, short>* required_version = &(*vlist)[required_vname];
		std::pair<short, short>* current_version = getVersion();

		// メジャーバージョン違い
		if(current_version->first != required_version->first) {
			HARKIO_SET_ERROR(std::string("Major Version Mismatch:")
			+ std::string(" Version:") + getVersionStr());
//			+ typeid(derive).name + string(" Version:") + getVersionStr());
			return MISMATCH;
		}
		//　マイナーバージョン違い
		if(current_version->second < required_version->second) {
			HARKIO_SET_ERROR(std::string("Major Version Mismatch:"));
//			+ typeid(derive).name + string(" Version:") + getVersionStr());
			return MISMATCH;
		}
		//　安全なマイナーバージョン違い
		if(current_version->second > required_version->second) {
			HARKIO_SET_ERROR(std::string("Major Version Safe Match"));
//			+ typeid(derive).name + string(" Version:") + getVersionStr());
			return SAFE_MATCH;
		}
		// 完全一致
		return PERFECT_MATCH;
	}

public:
	virtual ~Chunk(){}
	//! サイズヘッダーのサイズを含む自身のクラスのサイズを取得する関数。親の階層がsetSizeする時に呼ばれる
	virtual int getClassSize() = 0;
	//! 基底クラスからの呼び出しで派生クラスの識別子を取得する
	virtual const char* getIdentifierFromBase() = 0;
	//! サイズフィールドを設定する関数
	virtual void setSize() = 0;
	//! ファイルストリームからこのクラスのデータを読みとる
	virtual void readFile(std::ifstream &fs) = 0;
	//! ファイルストリームへこのクラスのデータを書き込む
	virtual void writeFile(std::ofstream &fs) = 0;
	//! サイズフィールドを取得する
	int getSize();
	//! インスタンスのバージョンを文字列で取得
	std::string getVersionStr() {
		std::ostringstream ss;
		ss << version.first << "." << version.second;
		return ss.str();
	}
	//! インスタンスのバージョン番号を取得
	std::pair<short, short>* getVersion() {
		return &version;
	}
};

/**********************************************************************/
//! FMT(Format)サブチャンククラス
/*!
 * HGTF Dataに必要なサブチャンク
 * NMIC,NFS,NFFT,マイク座標などのデータを保持する
 */
/**********************************************************************/
class FMT : public Chunk {
	friend class HGTF;
public:
	//! マイク座標の定義する型
	typedef std::vector<MIC_TYPE> MIC_LIST_TYPE;
	static const char *identifier;	//!< 識別子
	//! フォーマットＩＤ
	/*! FMTの次に連続するサブチャンクの種類を決めるID
	 */
	typedef enum {UNDEFINE=0, M2PG, GTF, SM} FORMAT_ID ;

	//! FMT のバージョン一覧
	typedef enum {
		V1_0 = 0, DELIMITER
	} VNAME;

private:
	//! \name フィールドデータ
	//\{
	FORMAT_ID format_id;   //!< このチャンクに連続するチャンク種類ID
	int nMic;             //!< マイクID
	int nFFT;             //!< NFFT
	int nIR;              //!< インパルスレスポンス長
	int nFS	;             //!< NFS
	MIC_LIST_TYPE mic;   //!< マイク位置の座標
	//\}

private:
	//! 書き込む前にデータが揃っているか確認する
	bool isComplete();

	//! サイズフィールドを設定する関数
	void setSize();

	//! サイズヘッダーのサイズを含む自身のクラスのサイズを取得する関数。親の階層がsetSizeする時に呼ばれる
	int getClassSize();

	//! ファイルストリームからこのクラスのデータを読みとる
	void readFile(std::ifstream &fs);

	//! ファイルストリームへこのクラスのデータを書き込む
	void writeFile(std::ofstream &fs);

	//! 基底クラスからの呼び出しで派生クラスの識別子を取得する
	const char* getIdentifierFromBase() {
		return getIdentifier();
	}
public:
	FMT() {} // ファイル読み込み用
	//! FMT コンストラクタ
	/*! \param vname FTM初期作成時の必須情報
	 */
	FMT(FMT::VNAME vname, FORMAT_ID format_id, int nMic, int nFFT, int nIR, int nFS) { // 新規作成用
		this->format_id = format_id;
		this->nMic      = nMic;
		this->nFFT      = nFFT;
		this->nIR       = nIR;
		this->nFS       = nFS;

		// バージョン名から、対応するバージョン値を設定する
		if (vname < 0 || vname >= DELIMITER) {
			HARKIO_THROW_ERROR("Invalid version ID");
		}
		std::vector<std::pair<short, short> > *vlist = getVersions();
		version = (*vlist)[vname];
	}
	~FMT() {
	}
	//! このクラスの識別子を取得する
	static const char *getIdentifier() {
		// 識別子
		static const char *identifier=NULL;
		if(!identifier) {
			identifier ="fmt ";
		}
		return identifier;
	}
	//! Dataのバージョン一覧を取得する
	static std::vector<std::pair<short, short> > *getVersions() {
		static std::vector<std::pair<short, short> > vlist;
		if (!vlist.size()) {
			vlist.resize(DELIMITER);
			vlist[V1_0] = std::pair<short, short>(1, 0);
		}
		return &vlist;
	}
	//! バージョン名からバージョン番号を取得
	static std::pair<short, short>* getVersionByName(VNAME vname) {
		if (vname < 0 || vname >= DELIMITER) {
			HARKIO_THROW_ERROR("Invalid version ID");
		}
		std::vector<std::pair<short, short> > *vlist = getVersions();
		return &((*vlist)[vname]);
	}
	//! 引数のバージョン名から、バージョンチェックを行う
	Chunk::VERSION_CHECK checkVersion(VNAME required_vname) {
		return checkVersionCommon<FMT>(this, required_vname);
	}
	//! フォーマットIDを取得する
	FORMAT_ID getFormatId();
	//! マイク座標のリストを取得する
	MIC_LIST_TYPE *getMic();
	int getNMic(); //!< NMIC取得
	int getNFFT(); //!< NFFT取得
	int getNIR();    //!< NIR(インパルスレスポンスの長さ)取得
	int getNFS();  //!< NFS取得
	//! マイク座標の配列を作成し、配列サイズも決める。
	MIC_LIST_TYPE *newMic();
};


/**********************************************************************/
//! M2PGクラス
/*!
 * HGTFデータに属するM2PG(Music 2D Polar-Grid)サブチャンク
 */
/**********************************************************************/
class M2PG : public Chunk{
	friend class HGTF;
public:
	//! 音源位置座表
	/*! 音源位置座標配列の型。src[nd][nr]でアクセスできる
	 */
	typedef std::vector<std::vector<MIC_TYPE> > SRC_LIST_TYPE;

	//! 伝達関数
	/*! 伝達関数配列の型。tf[nd][nr][nmic][nfft]でアクセスできる
	 */
	typedef std::vector<std::vector<std::vector<std::vector<TF_TYPE> > > > TF_LIST_TYPE;

	//! M2PG のバージョン一覧
	typedef enum {
		V1_0 = 0, DELIMITER
	} VNAME;

protected:
	FMT *fmt; // コンストラクション後に参照するFMT Chunkへの依存ポインタ
	//! \name フィールドデータ
	//\{
	int nd;             //!< 水平方向の要素数
	int nr;             //!< 距離の要素数
	SRC_LIST_TYPE src; //!< 音源位置座標
	TF_LIST_TYPE  tf;   //!< 伝達関数
	//\}

protected:
	//! 書き込む前にデータが揃っているか確認する
	bool isComplete();
	//! サイズフィールドを設定する関数
	void setSize();
	//! サイズヘッダーのサイズを含む自身のクラスのサイズを取得する関数。親の階層がsetSizeする時に呼ばれる
	int getClassSize();
	//! ファイルストリームからこのクラスのデータを読みとる
	void readFile(std::ifstream &ifs);
	//! ファイルストリームへこのクラスのデータを書き込む
	void writeFile(std::ofstream &fs);
	//! 基底クラスからの呼び出しで派生クラスの識別子を取得する
	const char* getIdentifierFromBase() {
		return getIdentifier();
	}
	//! 音源位置座標の配列サイズを決める
	SRC_LIST_TYPE *newSrc();
	//! 伝達関数の配列サイズを決める
	TF_LIST_TYPE *newTF();
public:
	//! ファイル読み込み時用のコンストラクタ
	M2PG(FMT *fmt) {
		this->fmt = fmt;
	}
	//! ファイル書き込み時用のコンストラクタ
	M2PG(M2PG::VNAME vname, FMT *fmt, int nd, int nr) {
		this->fmt = fmt;
		this->nd = nd;
		this->nr = nr;

		// バージョン名から、対応するバージョン値を設定する
		if (vname < 0 || vname >= DELIMITER) {
			HARKIO_THROW_ERROR("Invalid version ID");
		}
		std::vector<std::pair<short, short> > *vlist = getVersions();
		version = (*vlist)[vname];

		// 配列サイズを決める
		newSrc();
		newTF();
	}
	~M2PG() {
	}

	//! このクラスの識別子を取得する
	static const char *getIdentifier() {
		// 識別子
		static const char *identifier=NULL;
		if(!identifier) {
			identifier ="M2PG";
		}
		return identifier;
	}
	//! Dataのバージョン一覧を取得する
	static std::vector<std::pair<short, short> > *getVersions() {
		static std::vector<std::pair<short, short> > vlist;
		if (!vlist.size()) {
			vlist.resize(DELIMITER);
			vlist[V1_0] = std::pair<short, short>(1, 0);
		}
		return &vlist;
	}
	//! バージョン名からバージョン番号を取得
	static std::pair<short, short>* getVersionByName(VNAME vname) {
		if (vname < 0 || vname >= DELIMITER) {
			HARKIO_THROW_ERROR("Invalid version ID");
		}
		std::vector<std::pair<short, short> > *vlist = getVersions();
		return &((*vlist)[vname]);
	}
	//! 引数のバージョン名から、バージョンチェックを行う
	Chunk::VERSION_CHECK checkVersion(VNAME required_vname) {
		return Chunk::checkVersionCommon<M2PG>(this, required_vname);
	}

	int getND(); //!< ND取得
	int getNR(); //!< NR取得
	//! 音源位置座標の配列を取得する
	SRC_LIST_TYPE* getSrc();
	//! 伝達関数の配列を取得する
	TF_LIST_TYPE *getTF();
};

/**********************************************************************/
//! GTFクラス
/*!
 * HGTFデータに属するGTF(GSS Transfer Function)サブチャンク
 */
/**********************************************************************/
class GTF : public Chunk{
	friend class HGTF;
public:
	//! 音源位置座表
	/*! 音源位置座標配列の型。src[ns]でアクセスできる
	 */
	typedef std::vector<MIC_TYPE > SRC_LIST_TYPE;

	//! 伝達関数
	/*! 伝達関数配列の型。tf[ns][nmic][nf]でアクセスできる
	 */
	typedef std::vector<std::vector<std::vector<TF_TYPE> > > TF_LIST_TYPE;

	//! GTF のバージョン一覧
	typedef enum {
		V1_0 = 0, DELIMITER
	} VNAME;

protected:
	FMT *fmt; // コンストラクション後に参照するFMT Chunkへの依存ポインタ
	//! \name フィールドデータ
	//\{
	int n_src;             //!< 音源位置数
	SRC_LIST_TYPE src;  //!< 音源位置の座標
	int n_fi;             //!< 周波数ライン数
	std::vector<int> fi;   //!< 伝達関数のNFインデックスから周波数のマップ
	TF_LIST_TYPE  tf;   //!< 伝達関数
	//\}

	//! 音源位置座標の配列サイズを決める
	SRC_LIST_TYPE *newSrc();
	//! 伝達関数の配列サイズを決める
	TF_LIST_TYPE *newTF();
	//! 伝達関数のNFインデックスから周波数のマップ への配列サイズを決める
	std::vector<int> *newFI();

	protected:
	//! 書き込む前にデータが揃っているか確認する
	bool isComplete();
	//! サイズフィールドを設定する関数
	void setSize();
	//! サイズヘッダーのサイズを含む自身のクラスのサイズを取得する関数。親の階層がsetSizeする時に呼ばれる
	int getClassSize();
	//! ファイルストリームからこのクラスのデータを読みとる
	void readFile(std::ifstream &ifs);
	//! ファイルストリームへこのクラスのデータを書き込む
	void writeFile(std::ofstream &fs);
	//! 基底クラスからの呼び出しで派生クラスの識別子を取得する
	const char* getIdentifierFromBase() {
		return getIdentifier();
	}

public:
	//! ファイル読み込み時用のコンストラクタ
	GTF(FMT *fmt) {
		this->fmt = fmt;
	}
	//! ファイル書き込み時用のコンストラクタ
	GTF(GTF::VNAME vname, FMT *fmt, int n_src, int n_fi) {
		this->fmt = fmt;
		this->n_src  = n_src; // 音源位置の座標数
		this->n_fi  = n_fi; // 配列へのインデックスから周波数へのマップ数

		// バージョン名から、対応するバージョン値を設定する
		if (vname < 0 || vname >= DELIMITER) {
			HARKIO_THROW_ERROR("Invalid version ID");
		}
		std::vector<std::pair<short, short> > *vlist = getVersions();
		version = (*vlist)[vname];

		// 配列のサイズを決める
		newSrc();
		newTF();
		newFI();
	}
	~GTF() {
	}
	//! このクラスの識別子を取得する
	static const char *getIdentifier() {
		// 識別子
		static const char *identifier=NULL;
		if(!identifier) {
			identifier ="GTF ";
		}
		return identifier;
	}
	//! Dataのバージョン一覧を取得する
	static std::vector<std::pair<short, short> > *getVersions() {
		static std::vector<std::pair<short, short> > vlist;
		if (!vlist.size()) {
			vlist.resize(DELIMITER);
			vlist[V1_0] = std::pair<short, short>(1, 0);
		}
		return &vlist;
	}
	//! バージョン名からバージョン番号を取得
	static std::pair<short, short>* getVersionByName(VNAME vname) {
		if (vname < 0 || vname >= DELIMITER) {
			HARKIO_THROW_ERROR("Invalid version ID");
		}
		std::vector<std::pair<short, short> > *vlist = getVersions();
		return &((*vlist)[vname]);
	}
	//! 引数のバージョン名から、バージョンチェックを行う
	Chunk::VERSION_CHECK checkVersion(VNAME required_vname) {
		return Chunk::checkVersionCommon<GTF>(this, required_vname);
	}

	//! サイズフィールドを取得する
	int getNSrc(); //!< NSrc取得
	int getNFI(); //!< NFI取得
	//! 音源位置座標の配列を取得する
	SRC_LIST_TYPE* getSrc();
	//! 伝達関数の配列を取得する
	TF_LIST_TYPE *getTF();
	//! 伝達関数のFIインデックスから周波数のマップ の配列を取得
	std::vector<int>*getFI();
};
/**********************************************************************/
//! SMクラス
/*!
 * HGTFデータに属するSM(Separation Matrix)サブチャンク
 */
/**********************************************************************/
class SM : public Chunk{
	friend class HGTF;
public:
	//! 音源位置座表
	/*! 音源位置座標配列の型。src[ns]でアクセスできる
	 */
	typedef std::vector<MIC_TYPE > SRC_LIST_TYPE;

	//! 分離行列
	/*! 分離行列配列の型。tf[ns][nmic][nf]でアクセスできる
	 */
	typedef std::vector<std::vector<std::vector<TF_TYPE> > > SM_LIST_TYPE;

	//! SM のバージョン一覧
	typedef enum {
		V1_0 = 0, DELIMITER
	} VNAME;

protected:
	FMT *fmt; // コンストラクション後に参照するFMT Chunkへの依存ポインタ
	//! \name フィールドデータ
	//\{
	int n_src;             //!< 音源位置数
	SRC_LIST_TYPE src;  //!< 音源位置の座標
	int n_fi;             //!< 周波数ライン数
	std::vector<int> fi;//!< 分離行列のNFインデックスから周波数のマップ
	SM_LIST_TYPE  sm;   //!< 分離行列
	//\}

	//! 音源位置座標の配列サイズを決める
	SRC_LIST_TYPE *newSrc();
	//! 分離行列の配列サイズを決める
	SM_LIST_TYPE *newSM();
	//! 分離行列のNFインデックスから周波数のマップ への配列サイズを決める
	std::vector<int> *newFI();

	protected:
	//! 書き込む前にデータが揃っているか確認する
	bool isComplete();
	//! サイズフィールドを設定する関数
	void setSize();
	//! サイズヘッダーのサイズを含む自身のクラスのサイズを取得する関数。親の階層がsetSizeする時に呼ばれる
	int getClassSize();
	//! ファイルストリームからこのクラスのデータを読みとる
	void readFile(std::ifstream &ifs);
	//! ファイルストリームへこのクラスのデータを書き込む
	void writeFile(std::ofstream &fs);
	//! 基底クラスからの呼び出しで派生クラスの識別子を取得する
	const char* getIdentifierFromBase() {
		return getIdentifier();
	}
public:
	//! ファイル読み込み時用のコンストラクタ
	SM(FMT *fmt) {
		this->fmt = fmt;
	}
	//! ファイル書き込み時用のコンストラクタ
	SM(SM::VNAME vname, FMT *fmt, int n_src, int n_fi) {
		this->fmt = fmt;
		this->n_src  = n_src; // 音源位置の座標数
		this->n_fi  = n_fi; // 配列へのインデックスから周波数へのマップ数

		// バージョン名から、対応するバージョン値を設定する
		if (vname < 0 || vname >= DELIMITER) {
			HARKIO_THROW_ERROR("Invalid version ID");
		}
		std::vector<std::pair<short, short> > *vlist = getVersions();
		version = (*vlist)[vname];

		// 配列のサイズを決める
		newSrc();
		newSM();
		newFI();
	}
	~SM() {
	}
	//! このクラスの識別子を取得する
	static const char *getIdentifier() {
		// 識別子
		static const char *identifier=NULL;
		if(!identifier) {
			identifier ="SM  ";
		}
		return identifier;
	}
	//! Dataのバージョン一覧を取得する
	static std::vector<std::pair<short, short> > *getVersions() {
		static std::vector<std::pair<short, short> > vlist;
		if (!vlist.size()) {
			vlist.resize(DELIMITER);
			vlist[V1_0] = std::pair<short, short>(1, 0);
		}
		return &vlist;
	}
	//! バージョン名からバージョン番号を取得
	static std::pair<short, short>* getVersionByName(VNAME vname) {
		if (vname < 0 || vname >= DELIMITER) {
			HARKIO_THROW_ERROR("Invalid version ID");
		}
		std::vector<std::pair<short, short> > *vlist = getVersions();
		return &((*vlist)[vname]);
	}
	//! 引数のバージョン名から、バージョンチェックを行う
	Chunk::VERSION_CHECK checkVersion(VNAME required_vname) {
		return Chunk::checkVersionCommon<SM>(this, required_vname);
	}

	//! サイズフィールドを取得する
	int getNSrc(); //!< NSrc取得
	int getNFI(); //!< NR取得
	//! 音源位置座標の配列を取得する
	SRC_LIST_TYPE* getSrc();
	//! 分離行列の配列を取得する
	SM_LIST_TYPE *getSM();
	//! 伝達関数のNFインデックスから周波数のマップ の配列を取得
	std::vector<int>*getFI();
};

/**********************************************************************/
//! SPECクラス
/*!
 * HGTFデータに属するSPECサブチャンク
 */
/**********************************************************************/
class SPEC : public Chunk{
	friend class HGTF;
public:
	//! SPEC のバージョン一覧
	typedef enum {
		V1_0 = 0, DELIMITER
	} VNAME;

protected:
	//! \name フィールドデータ
	//\{
	char date[11];				//!< 作成日(2009/05/19)
	char location[128];			//!< 場所 (スマートルーム)
	char robot[128];			//!< ロボット(ASIMO ver 2.5#4 サーボオン)
	char array_type[128];		//!< アレイのタイプ(8本鉢巻,8本キューブ)
	char creator[128];			//!< 作成者(山本 俊一)
	char memo[1024];			//!< 備忘録(MUSICのTF作成パラメータなど)
	//\}

protected:
	//! 書き込む前にデータが揃っているか確認する
	bool isComplete();
	//! サイズフィールドを設定する関数
	void setSize();
	//! サイズヘッダーのサイズを含む自身のクラスのサイズを取得する関数。親の階層がsetSizeする時に呼ばれる
	int getClassSize();
	//! ファイルストリームからこのクラスのデータを読みとる
	void readFile(std::ifstream &ifs);
	//! ファイルストリームへこのクラスのデータを書き込む
	void writeFile(std::ofstream &fs);
	//! 基底クラスからの呼び出しで派生クラスの識別子を取得する
	const char* getIdentifierFromBase() {
		return getIdentifier();
	}
public:
	//! ファイル読み込み時用コンストラクタ
	SPEC() {
	}
	//! ファイル書き込み時用コンストラクタ
	SPEC(VNAME vname) {
		// バージョン名から、対応するバージョン値を設定する
		if (vname < 0 || vname >= DELIMITER) {
			HARKIO_THROW_ERROR("Invalid version ID");
		}
		std::vector<std::pair<short, short> > *vlist = getVersions();
		version = (*vlist)[vname];
		strcpy(date, "");
		strcpy(location, "");
		strcpy(robot, "");
		strcpy(array_type, "");
		strcpy(creator, "");
		strcpy(memo, "");
	}
	~SPEC() {
	}
	//! このクラスの識別子を取得する
	static const char *getIdentifier() {
		// 識別子
		static const char *identifier=NULL;
		if(!identifier) {
			identifier ="SPEC";
		}
		return identifier;
	}
	//! Dataのバージョン一覧を取得する
	static std::vector<std::pair<short, short> > *getVersions() {
		static std::vector<std::pair<short, short> > vlist;
		if (!vlist.size()) {
			vlist.resize(DELIMITER);
			vlist[V1_0] = std::pair<short, short>(1, 0);
		}
		return &vlist;
	}
	//! バージョン名からバージョン番号を取得
	static std::pair<short, short>* getVersionByName(VNAME vname) {
		if (vname < 0 || vname >= DELIMITER) {
			HARKIO_THROW_ERROR("Invalid version ID");
		}
		std::vector<std::pair<short, short> > *vlist = getVersions();
		return &((*vlist)[vname]);
	}
	//! 引数のバージョン名から、バージョンチェックを行う
	Chunk::VERSION_CHECK checkVersion(VNAME required_vname) {
		return Chunk::checkVersionCommon<SPEC>(this, required_vname);
	}

	//! フィールドデータの設定
	void setDate(const char *date) {
		strncpy(this->date, date, sizeof(this->date));
	}
	void setLocation(const char *location) {
		strncpy(this->location, location, sizeof(this->location));
	}
	void setRobot(const char *robot) {
		strncpy(this->robot, robot, sizeof(this->robot));
	}
	void setArrayType(const char *array_type) {
		strncpy(this->array_type, array_type, sizeof(this->array_type));
	}
	void setCreator(const char *creator) {
		strncpy(this->creator, creator, sizeof(this->creator));
	}
	void setMemo(const char *memo) {
		strncpy(this->memo, memo, sizeof(this->memo));
	}

	//! フィールドデータの取得
	const char* getDate() {
		return date;
	}
	const char* getLocation() {
		return location;
	}
	const char* getRobot() {
		return robot;
	}
	const char* getArrayType() {
		return array_type;
	}
	const char* getCreator() {
		return creator;
	}
	const char* getMemo() {
		return memo;
	}
};

}// namespace

#endif


